import os
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Any, Dict, List

from cognite.client._api_client import APIClient
from cognite.client.data_classes.model_hosting.versions import (
    ModelArtifactList,
    ModelVersion,
    ModelVersionList,
    ModelVersionLog,
)


class EmptyArtifactsDirectory(Exception):
    pass


class ModelVersionsAPI(APIClient):
    def create_model_version(
        self, model_name: str, version_name: str, source_package_id: int, description: str = None, metadata: Dict = None
    ) -> ModelVersion:
        """Create a model version without deploying it.

        Then you can optionally upload artifacts to the model version and later deploy it.

        Args:
            model_name (str): Create the version on the model with this name.
            version_name (str): Name of the the model version.
            source_package_id (int): Use the source package with this id. The source package must have an available
                predict operation.
            description (str):  Description of model version
            metadata (Dict[str, Any]):  Metadata about model version

        Returns:
            ModelVersion: The created model version.
        """
        url = "/modelhosting/models/{}/versions".format(model_name)
        body = {
            "name": version_name,
            "description": description or "",
            "sourcePackageId": source_package_id,
            "metadata": metadata or {},
        }
        res = self._post(url, json=body)
        return ModelVersion._load(res.json())

    def deploy_awaiting_model_version(self, model_name: str, version_name: str) -> ModelVersion:
        """Deploy an already created model version awaiting manual deployment.

        The model version must have status AWAITING_MANUAL_DEPLOYMENT in order for this to work.

        Args:
            model_name (str): The name of the model containing the version to deploy.
            version_name (str): The name of the model version to deploy.
        Returns:
            ModelVersion: The deployed model version.
        """
        url = "/modelhosting/models/{}/versions/{}/deploy".format(model_name, version_name)
        res = self._post(url, json={})
        return ModelVersion._load(res.json())

    def deploy_model_version(
        self,
        model_name: str,
        version_name: str,
        source_package_id: int,
        artifacts_directory: str = None,
        description: str = None,
        metadata: Dict = None,
    ) -> ModelVersion:
        """This will create and deploy a model version.

        If artifacts_directory is specified, it will traverse that directory recursively and
        upload all artifacts in that directory before deploying.

        Args:
            model_name (str): Create the version on the model with this name.
            version_name (str): Name of the the model version.
            source_package_id (int): Use the source package with this id. The source package must have an available
                predict operation.
            artifacts_directory (str, optional): Absolute path of directory containing artifacts.
            description (str, optional):  Description of model version
            metadata (Dict[str, Any], optional):  Metadata about model version

        Returns:
            ModelVersion: The created model version.
        """
        model_version = self.create_model_version(model_name, version_name, source_package_id, description, metadata)
        if artifacts_directory:
            self.upload_artifacts_from_directory(model_name, model_version.name, directory=artifacts_directory)
        return self.deploy_awaiting_model_version(model_name, model_version.name)

    def list_model_versions(
        self, model_name: str, limit: int = None, cursor: str = None, autopaging: bool = False
    ) -> ModelVersionList:
        """Get all versions of a specific model.

        Args:
            model_name (str): Get versions for the model with this name.
            limit (int): Maximum number of model versions to return. Defaults to 250.
            cursor (str): Cursor to use to fetch next set of results.
            autopaging (bool): Whether or not to automatically page through all results. Will disregard limit.

        Returns:
            ModelVersionList: List of model versions
        """
        url = "/modelhosting/models/{}/versions".format(model_name)
        params = {"cursor": cursor, "limit": limit if autopaging is False else self._LIST_LIMIT}
        res = self._get(url, params=params)
        return ModelVersionList._load(res.json()["items"])

    def get_model_version(self, model_name: str, version_name: str) -> ModelVersion:
        """Get a specific model version by name.

        Args:
            model_name (str): Name of model which has the model version.
            version_name (str): Name of model version.

        Returns:
            ModelVersion: The requested model version
        """
        url = "/modelhosting/models/{}/versions/{}".format(model_name, version_name)
        res = self._get(url)
        return ModelVersion._load(res.json())

    def update_model_version(
        self, model_name: str, version_name: str, description: str = None, metadata: Dict[str, str] = None
    ) -> ModelVersion:
        """Update description or metadata on a model version.

        Args:
            model_name (str): Name of model containing the model version.
            version_name (str): Name of model version to update.
            description (str): New description.
            metadata (Dict[str, str]): New metadata

        Returns:
            ModelVersion: The updated model version.
        """
        url = "/modelhosting/models/{}/versions/{}/update".format(model_name, version_name)
        body = {}
        if description:
            body.update({"description": {"set": description}})
        if metadata:
            body.update({"metadata": {"set": metadata}})

        res = self._post(url, json=body)
        return ModelVersion._load(res.json())

    def deprecate_model_version(self, model_name: str, version_name: str) -> ModelVersion:
        """Deprecate a model version

        Args:
            model_name (str): Name of model
            version_name (str): name of model version to deprecate

        Returns:
            ModelVersion: The deprecated model version
        """
        url = "/modelhosting/models/{}/versions/{}/deprecate".format(model_name, version_name)
        res = self._post(url)
        return ModelVersion._load(res.json())

    def delete_model_version(self, model_name: str, version_name: str) -> None:
        """Delete a model version by id.

        Args:
            model_name (str): Name of model which has the model version.
            version_name (str): Name of model version.

        Returns:
            None
        """
        url = "/modelhosting/models/{}/versions/{}".format(model_name, version_name)
        self._delete(url)

    def list_artifacts(self, model_name: str, version_name: str) -> ModelArtifactList:
        """List the artifacts associated with the specified model version.

        Args:
            model_name (str): Name of model
            version_name (str): Name of model version to get artifacts for

        Returns:
            ModelArtifactList: List of artifacts
        """
        url = "/modelhosting/models/{}/versions/{}/artifacts".format(model_name, version_name)
        res = self._get(url)
        return ModelArtifactList._load(res.json()["items"])

    def download_artifact(self, model_name: str, version_name: str, artifact_name: str, directory: str = None) -> None:
        """Download an artifact to a directory. Defaults to current working directory.

        Args:
            model_name (str): Name of model
            version_name (str): Name of model version.
            artifact_name (str): Name of artifact.
            directory (str): Directory to place artifact in. Defaults to current working directory.

        Returns:
            None
        """
        directory = directory or os.getcwd()
        file_path = os.path.join(directory, artifact_name)

        url = "/modelhosting/models/{}/versions/{}/artifacts/{}".format(model_name, version_name, artifact_name)
        download_url = self._get(url).json()["downloadUrl"]
        with open(file_path, "wb") as fh:
            response = self._http_client.request("GET", download_url).content
            fh.write(response)

    def upload_artifact_from_file(self, model_name: str, version_name: str, artifact_name: str, file_path: str) -> None:
        """Upload an artifact to a model version.

        The model version must have status AWAITING_MANUAL_DEPLOYMENT in order for this to work.

        Args:
            model_name (str): The name of the model.
            version_name (str): The name of the model version to upload the artifacts to.
            artifact_name (str): The name of the artifact.
            file_path (str): The local path of the artifact.
        Returns:
            None
        """
        url = "/modelhosting/models/{}/versions/{}/artifacts/upload".format(model_name, version_name)
        body = {"name": artifact_name}
        res = self._post(url, json=body)
        upload_url = res.json()["uploadUrl"]
        self._upload_file(upload_url, file_path)

    def upload_artifacts_from_directory(self, model_name: str, version_name: str, directory: str) -> None:
        """Upload all files in directory recursively.

        Args:
            model_name (str): The name of the model.
            version_name (str): The name of the model version to upload the artifacts to.
            directory (str): Absolute path of directory to upload artifacts from.
        Returns:
            None
        """
        upload_tasks = []
        for root, dirs, files in os.walk(directory):
            for file_name in files:
                file_path = os.path.join(root, file_name)
                full_file_name = os.path.relpath(file_path, directory)
                upload_tasks.append((model_name, version_name, full_file_name, file_path))

        if len(upload_tasks) == 0:
            raise EmptyArtifactsDirectory("Artifacts directory is empty.")
        self._execute_tasks_concurrently(self.upload_artifact_from_file, upload_tasks)

    @staticmethod
    def _execute_tasks_concurrently(func, tasks):
        with ThreadPoolExecutor(16) as p:
            futures = [p.submit(func, *task) for task in tasks]
            return [future.result() for future in futures]

    def _upload_file(self, upload_url, file_path):
        with open(file_path, "rb") as fh:
            mydata = fh.read()
            response = self._http_client.request("PUT", upload_url, data=mydata)
        return response

    def get_logs(self, model_name: str, version_name: str, log_type: str = None) -> ModelVersionLog:
        """Get logs for prediction and/or training routine of a specific model version.

        Args:
            model_name (str): Name of model.
            version_name (str): Name of model version to get logs for.
            log_type (str): Which routine to get logs from. Must be 'train’, 'predict’, or ‘both’. Defaults to 'both'.

        Returns:
            ModelVersionLog: An object containing the requested logs.
        """
        url = "/modelhosting/models/{}/versions/{}/log".format(model_name, version_name)
        params = {"logType": log_type}
        res = self._get(url, params=params)
        return ModelVersionLog._load(res.json())



from typing import List, Optional

from cognite.client.data_classes._base import *


class TemplateGroup(CogniteResource):
    """A template group is a high level concept encapsulating a schema and a set of template instances. It also has query capability support.

    Template groups are versioned, so there can be multiple template groups with the same external ID.
    The versioning is happening automatically whenever a template groups is changed.

    GraphQL schema definition language is used as the language to describe the structure of the templates and data types.

    Args:
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
        description (str): The description of the template groups.
        owners (List[str]): The list of owners for the template groups.
    """

    def __init__(
        self,
        external_id: str = None,
        description: str = None,
        owners: Optional[List[str]] = None,
        created_time: int = None,
        last_updated_time: int = None,
        cognite_client=None,
    ):
        self.external_id = external_id
        self.description = description
        self.owners = owners
        self.created_time = created_time
        self.last_updated_time = last_updated_time
        self._cognite_client = cognite_client


class TemplateGroupList(CogniteResourceList):
    _RESOURCE = TemplateGroup
    _UPDATE = CogniteUpdate


class TemplateGroupVersion(CogniteResource):
    """
    A Template Group Version supports specifying different conflict modes, which is used when an existing schema already exists.

    Patch -> It diffs the new schema with the old schema and fails if there are breaking changes.
    Update -> It sets the new schema as schema of a new version.
    Force -> It ignores breaking changes and replaces the old schema with the new schema.
    The default mode is "patch".

    Args:
        schema (str): The GraphQL schema.
        version (int): Incremented by the server whenever the schema of a template groups changes.
        conflict_mode (str): Can be set to 'Patch', 'Update' or 'Force'.
    """

    def __init__(
        self,
        schema: str = None,
        version: int = None,
        conflict_mode: str = None,
        created_time: int = None,
        last_updated_time: int = None,
        cognite_client=None,
    ):
        self.schema = schema
        self.version = version
        self.conflict_mode = conflict_mode
        self.created_time = created_time
        self.last_updated_time = last_updated_time
        self._cognite_client = cognite_client


class TemplateGroupVersionList(CogniteResourceList):
    _RESOURCE = TemplateGroupVersion
    _UPDATE = CogniteUpdate


class ConstantResolver(CogniteResource):
    """Resolves a field to a constant value. The value can be of any supported JSON type.

    Args:
        value (any): The value of the field.
    """

    def __init__(self, value: any = None, cognite_client=None):
        self.type = "constant"
        self.value = value
        self._cognite_client = cognite_client


class RawResolver(CogniteResource):
    """Resolves a field to a RAW column.

    Args:
        db_name (str): The database name.
        table_name (str): The table name.
        row_key (str): The row key.
        column_name (str): The column to fetch the value from.
    """

    def __init__(
        self,
        db_name: str = None,
        table_name: str = None,
        row_key: str = None,
        column_name: str = None,
        cognite_client=None,
    ):
        self.type = "raw"
        self.db_name = db_name
        self.table_name = table_name
        self.row_key = row_key
        self.column_name = column_name
        self._cognite_client = cognite_client


class SyntheticTimeSeriesResolver(CogniteResource):
    """Resolves a field of type 'SyntheticTimeSeries' to a Synthetic Time Series.

    Args:
        expression (str): The synthetic time series expression. See this for syntax https://docs.cognite.com/api/v1/#tag/Synthetic-Time-Series.
        name (Optional[str]): The name of the Time Series.
        description (Optional[str]): The description for the Time Series.
        metadata (Optional[Dict[str, str]]): Specifies metadata for the Time Series.
        is_step (Optional[bool]): Specifies if the synthetic time series is step based.
        is_string (Optional[bool]): Specifies if the synthetic time series returned contains string values.
        unit (Optional[str]): The unit of the time series.
    """

    def __init__(
        self,
        expression: str = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        metadata: Optional[Dict[str, str]] = None,
        is_step: Optional[bool] = None,
        is_string: Optional[bool] = None,
        unit: Optional[str] = None,
        cognite_client=None,
    ):
        self.type = "syntheticTimeSeries"
        self.expression = expression
        self.name = name
        self.description = description
        self.metadata = metadata
        self.is_step = is_step
        self.is_string = is_string
        self.unit = unit
        self._cognite_client = cognite_client


FieldResolvers = Union[ConstantResolver, RawResolver, SyntheticTimeSeriesResolver, str]


class TemplateInstance(CogniteResource):
    """A template instance that implements a template by specifying a resolver per field.

    Args:
        external_id (str): The id of the template instance.
        template_name (str): The template name to implement.
        field_resolvers (Dict[str, FieldResolvers]): A set of field resolvers where the dictionary key correspond to the field name.
        data_set_id (int): The id of the dataset this instance belongs to.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
    """

    def __init__(
        self,
        external_id: str = None,
        template_name: str = None,
        field_resolvers: Dict[str, FieldResolvers] = None,
        data_set_id: Optional[int] = None,
        created_time: int = None,
        last_updated_time: int = None,
        cognite_client=None,
    ):
        self.external_id = external_id
        self.template_name = template_name
        self.field_resolvers = field_resolvers
        self.data_set_id = data_set_id
        self.created_time = created_time
        self.last_updated_time = last_updated_time
        self._cognite_client = cognite_client

    field_resolver_mapper = {
        "constant": ConstantResolver,
        "syntheticTimeSeries": SyntheticTimeSeriesResolver,
        "raw": RawResolver,
    }

    def dump(self, camel_case: bool = False) -> Dict[str, Any]:
        """Dump the instance into a json serializable Python data type.

        Args:
            camel_case (bool): Use camelCase for attribute names. Defaults to False.

        Returns:
            Dict[str, Any]: A dictionary representation of the instance.
        """
        if camel_case:
            return {
                utils._auxiliary.to_camel_case(key): value
                if key != "field_resolvers"
                else TemplateInstance._encode_field_resolvers(value, camel_case=camel_case)
                for key, value in self.__dict__.items()
                if value not in EXCLUDE_VALUE and not key.startswith("_")
            }
        return {
            key: value
            if key != "field_resolvers"
            else TemplateInstance._encode_field_resolvers(value, camel_case=camel_case)
            for key, value in self.__dict__.items()
            if value not in EXCLUDE_VALUE and not key.startswith("_")
        }

    def _encode_field_resolvers(field_resolvers: Dict[str, FieldResolvers], camel_case):
        return {key: value.dump(camel_case=camel_case) for key, value in field_resolvers.items()}

    @classmethod
    def _load(cls, resource: Union[Dict, str], cognite_client=None):
        if isinstance(resource, str):
            return cls._load(json.loads(resource), cognite_client=cognite_client)
        elif isinstance(resource, Dict):
            instance = cls(cognite_client=cognite_client)
            for key, value in resource.items():
                snake_case_key = utils._auxiliary.to_snake_case(key)
                if hasattr(instance, snake_case_key):
                    if key == "fieldResolvers":
                        setattr(
                            instance,
                            snake_case_key,
                            {
                                key: TemplateInstance._field_resolver_load(field_resolver)
                                for key, field_resolver in value.items()
                            },
                        )
                    else:
                        setattr(instance, snake_case_key, value)
            return instance
        raise TypeError("Resource must be json str or Dict, not {}".format(type(resource)))

    def _field_resolver_load(resource: Union[Dict, str], cognite_client=None):
        return TemplateInstance.field_resolver_mapper[resource["type"]]._load(resource, cognite_client)


class GraphQlError(CogniteResource):
    def __init__(
        self, message: str = None, path: List[str] = None, locations: List[Dict[str, Any]] = None, cognite_client=None
    ):
        self.message = message
        self.path = path
        self.locations = locations
        self._cognite_client = cognite_client


class GraphQlResponse(CogniteResource):
    def __init__(self, data: any = None, errors: List[GraphQlError] = None, cognite_client=None):
        self.data = data
        self.errors = errors
        self._cognite_client = cognite_client


class TemplateInstanceList(CogniteResourceList):
    _RESOURCE = TemplateInstance
    _UPDATE = CogniteUpdate



import os
import re
from subprocess import check_call
from typing import Dict, List, Tuple

from cognite.client._api_client import APIClient
from cognite.client.data_classes.model_hosting.source_packages import (
    CreateSourcePackageResponse,
    SourcePackage,
    SourcePackageList,
)


class SourcePackagesAPI(APIClient):
    def upload_source_package(
        self,
        name: str,
        package_name: str,
        available_operations: List[str],
        runtime_version: str,
        description: str = None,
        metadata: Dict = None,
        file_path: str = None,
    ) -> CreateSourcePackageResponse:
        """Upload a source package to the model hosting environment.

        Args:
            name (str): Name of source package
            package_name (str): name of root package for model
            available_operations (List[str]): List of routines which this source package supports ["predict", "train"]
            runtime_version (str): Version of environment in which the source-package should run. Currently only 0.1.
            description (str): Description for source package
            metadata (Dict): User defined key value pair of additional information.
            file_path (str): File path of source package distribution. If not specified, a download url will be returned.

        Returns:
            CreateSourcePackageResponse: An response object containing Source package ID
            if file_path was specified. Else, both source package id and upload url.

        """
        body = {
            "name": name,
            "description": description or "",
            "packageName": package_name,
            "availableOperations": available_operations,
            "metadata": metadata or {},
            "runtimeVersion": runtime_version,
        }
        res = self._post("/modelhosting/models/sourcepackages", json=body).json()
        if file_path:
            self._upload_file(res["uploadUrl"], file_path)
            del res["uploadUrl"]
            return CreateSourcePackageResponse._load(res)
        return CreateSourcePackageResponse._load(res)

    def _get_model_py_files(self, path) -> List:
        files_containing_model_py = []
        for root, dirs, files in os.walk(path):
            if "model.py" in files:
                package_name = os.path.basename(root)
                file_path = os.path.join(root, "model.py")
                files_containing_model_py.append((package_name, file_path))
        return files_containing_model_py

    def _find_model_file_and_extract_details(self, package_directory: str) -> Tuple:
        num_of_eligible_model_py_files = 0
        for package_name, file_path in self._get_model_py_files(package_directory):
            with open(file_path, "r") as f:
                file_content = f.read()
                if re.search("class Model", file_content):
                    num_of_eligible_model_py_files += 1
                    model_package_name = package_name
                    model_file_content = file_content

        assert num_of_eligible_model_py_files != 0, "Could not locate a file named model.py containing a Model class"
        assert num_of_eligible_model_py_files == 1, "Multiple model.py files with a Model class in your source package"

        available_operations = []
        if re.search(r"def train\(", model_file_content):
            available_operations.append("TRAIN")
        if re.search(r"def predict\(", model_file_content):
            if re.search(r"def load\(", model_file_content):
                available_operations.append("PREDICT")
            else:
                raise AssertionError("Your Model class defines predict() but not load().")
        assert len(available_operations) > 0, "Your model does not define a train or a predict method"
        return model_package_name, available_operations

    def _build_distribution(self, package_path) -> str:
        check_call("cd {} && python setup.py sdist".format(package_path), shell=True)
        dist_directory = os.path.join(package_path, "dist")
        for file in os.listdir(dist_directory):
            if file.endswith(".tar.gz"):
                return os.path.join(dist_directory, file)

    def build_and_upload_source_package(
        self, name: str, runtime_version: str, package_directory: str, description: str = None, metadata: Dict = None
    ) -> CreateSourcePackageResponse:
        """Build a distribution for a source package and upload it to the model hosting environment.

        This method will recursively search through your package and infer available_operations as well as the package
        name.

        Args:
            name (str): Name of source package
            runtime_version (str): Version of environment in which the source-package should run. Currently only 0.1.
            description (str): Description for source package
            metadata (Dict): User defined key value pair of additional information.
            package_directory (str): Absolute path of directory containing your setup.py file.

        Returns:
            CreateSourcePackageResponse: An response object containing Source package ID
            if file_path was specified. Else, both source package id and upload url.
        """
        package_name, available_operations = self._find_model_file_and_extract_details(package_directory)
        tar_gz_path = self._build_distribution(package_directory)

        return self.upload_source_package(
            name=name,
            package_name=package_name,
            available_operations=available_operations,
            runtime_version=runtime_version,
            description=description,
            metadata=metadata,
            file_path=tar_gz_path,
        )

    def _upload_file(self, upload_url, file_path):
        with open(file_path, "rb") as fh:
            mydata = fh.read()
            response = self._http_client.request("PUT", upload_url, data=mydata, timeout=180)
        return response

    def list_source_packages(
        self, limit: int = None, cursor: str = None, autopaging: bool = False
    ) -> SourcePackageList:
        """List all model source packages.

        Args:
            limit (int): Maximum number of source_packages to return. Defaults to 250.
            cursor (str): Cursor to use to fetch next set of results.
            autopaging (bool): Whether or not to automatically page through all results. Will disregard limit.

        Returns:
            SourcePackageList: List of source packages.
        """
        params = {"cursor": cursor, "limit": limit if autopaging is False else self._LIST_LIMIT}
        res = self._get("/modelhosting/models/sourcepackages", params=params)
        return SourcePackageList._load(res.json()["items"])

    def get_source_package(self, id: int) -> SourcePackage:
        """Get source package by id.

        Args:
            id (int): Id of soure package to get.

        Returns:
            SourcePackage: The requested source package.
        """
        res = self._get("/modelhosting/models/sourcepackages/{}".format(id))
        return SourcePackage._load(res.json())

    def delete_source_package(self, id: int) -> None:
        """Delete source package by id.

        Args:
            id (int): Id of soure package to delete.

        Returns:
            None
        """
        self._delete("/modelhosting/models/sourcepackages/{}".format(id))

    def deprecate_source_package(self, id: int) -> SourcePackage:
        """Deprecate a source package by id.

        Args:
            id (int): Id of soure package to get.

        Returns:
            SourcePackage: The requested source package.
        """
        res = self._post("/modelhosting/models/sourcepackages/{}/deprecate".format(id), json={})
        return SourcePackage._load(res.json())

    def download_source_package_code(self, id: int, directory: str = None) -> None:
        """Download the tarball for a source package to a specified directory.

        Args:
            id (int): Id of source package.
            directory (str): Directory to put source package in. Defaults to current working directory.

        Returns:
            None
        """
        directory = directory or os.getcwd()
        file_path = os.path.join(directory, self.get_source_package(id).name + ".tar.gz")
        url = "/modelhosting/models/sourcepackages/{}/code".format(id)
        download_url = self._get(url).json()["downloadUrl"]
        with open(file_path, "wb") as fh:
            response = self._http_client.request("GET", download_url).content
            fh.write(response)

    def delete_source_package_code(self, id: int) -> None:
        """Delete the code/tarball for the source package from the cloud storage location.
        This will only work if the source package has been deprecated.

        Warning: This cannot be undone.

        Args:
            id (int): Id of the source package.

        Returns:
            None
        """
        self._delete("/modelhosting/models/sourcepackages/{}/code".format(id))
